@@ -1,6 +1,6 @@
 
 /*
- * $Id: forward.c,v 1.117 2006/09/30 21:10:48 hno Exp $
+ * $Id: forward.c,v 1.131.2.5 2008/07/18 00:47:48 hno Exp $
  *
  * DEBUG: section 17    Request Forwarding
  * AUTHOR: Duane Wessels
@@ -37,6 +37,7 @@
 #include "squid.h"
 
 #if LINUX_NETFILTER
+#include <linux/types.h>
 #include <linux/netfilter_ipv4.h>
 #endif
 #if LINUX_TPROXY
@@ -53,6 +54,7 @@ static PF fwdPeerClosed;
 static CNCB fwdConnectDone;
 static int fwdCheckRetry(FwdState * fwdState);
 static int fwdReforward(FwdState *);
+static void fwdRestart(void *);
 static void fwdStartFail(FwdState *);
 static void fwdLogReplyStatus(int tries, http_status status);
 static OBJH fwdStats;
@@ -100,15 +102,8 @@ fwdStateFree(FwdState * fwdState)
     fwdLog(fwdState);
 #endif
     if (e->store_status == STORE_PENDING) {
-	if (e->mem_obj->inmem_hi == 0) {
-	    assert(fwdState->err);
-	    errorAppendEntry(e, fwdState->err);
-	    fwdState->err = NULL;
-	} else {
-	    EBIT_CLR(e->flags, ENTRY_FWD_HDR_WAIT);
-	    storeComplete(e);
-	    storeReleaseRequest(e);
-	}
+	storeRequestFailed(e, fwdState->err);
+	fwdState->err = NULL;
     }
     if (EBIT_TEST(e->flags, ENTRY_DEFER_READ))
 	storeResetDefer(e);
@@ -120,7 +115,7 @@ fwdStateFree(FwdState * fwdState)
     fwdState->request = NULL;
     if (fwdState->err)
 	errorStateFree(fwdState->err);
-    storeClientUnregisterAbort(e);
+    storeUnregisterAbort(e);
     storeUnlockObject(e);
     fwdState->entry = NULL;
     sfd = fwdState->server_fd;
@@ -214,11 +209,11 @@ fwdServerClosed(int fd, void *data)
 	    }
 	}
 	/* use eventAdd to break potential call sequence loops and to slow things down a little */
-	eventAdd("fwdConnectStart", fwdConnectStart, fwdState, originserver ? 0.05 : 0.005, 0);
+	eventAdd("fwdRestart", fwdRestart, fwdState, originserver ? 0.05 : 0.005, 0);
 	return;
     }
     if (!fwdState->err && shutting_down)
-	fwdState->err = errorCon(ERR_SHUTTING_DOWN, HTTP_SERVICE_UNAVAILABLE, fwdState->request);
+	fwdState->err = errorCon(ERR_SHUTTING_DOWN, HTTP_GATEWAY_TIMEOUT, fwdState->request);
     fwdStateFree(fwdState);
 }
 
@@ -246,7 +241,7 @@ fwdNegotiateSSL(int fd, void *data)
 	    return;
 	default:
 	    debug(81, 1) ("fwdNegotiateSSL: Error negotiating SSL connection on FD %d: %s (%d/%d/%d)\n", fd, ERR_error_string(ERR_get_error(), NULL), ssl_error, ret, errno);
-	    err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE, request);
+	    err = errorCon(ERR_CONNECT_FAIL, HTTP_BAD_GATEWAY, request);
 #ifdef EPROTO
 	    err->xerrno = EPROTO;
 #else
@@ -319,6 +314,7 @@ fwdInitiateSSL(FwdState * fwdState)
     fd_table[fd].ssl = ssl;
     fd_table[fd].read_method = &ssl_read_method;
     fd_table[fd].write_method = &ssl_write_method;
+    fd_note(fd, "Negotiating SSL");
     fwdNegotiateSSL(fd, fwdState);
 }
 #endif
@@ -343,13 +339,13 @@ fwdConnectDone(int server_fd, int status, void *data)
 	    fwdState->flags.dont_retry = 1;
 	debug(17, 4) ("fwdConnectDone: Unknown host: %s\n",
 	    request->host);
-	err = errorCon(ERR_DNS_FAIL, HTTP_SERVICE_UNAVAILABLE, fwdState->request);
+	err = errorCon(ERR_DNS_FAIL, HTTP_GATEWAY_TIMEOUT, fwdState->request);
 	err->dnsserver_msg = xstrdup(dns_error_message);
 	fwdFail(fwdState, err);
 	comm_close(server_fd);
     } else if (status != COMM_OK) {
 	assert(fs);
-	err = errorCon(ERR_CONNECT_FAIL, HTTP_SERVICE_UNAVAILABLE, fwdState->request);
+	err = errorCon(ERR_CONNECT_FAIL, HTTP_GATEWAY_TIMEOUT, fwdState->request);
 	err->xerrno = errno;
 	fwdFail(fwdState, err);
 	if (fs->peer)
@@ -357,10 +353,6 @@ fwdConnectDone(int server_fd, int status, void *data)
 	comm_close(server_fd);
     } else {
 	debug(17, 3) ("fwdConnectDone: FD %d: '%s'\n", server_fd, storeUrl(fwdState->entry));
-	fd_note(server_fd, storeUrl(fwdState->entry));
-	fd_table[server_fd].uses++;
-	if (fd_table[server_fd].uses == 1 && fs->peer)
-	    peerConnectSucceded(fs->peer);
 #if USE_SSL
 	if ((fs->peer && fs->peer->use_ssl) ||
 	    (!fs->peer && request->protocol == PROTO_HTTPS)) {
@@ -397,11 +389,82 @@ fwdConnectTimeout(int fd, void *data)
     comm_close(fd);
 }
 
+typedef struct {
+    peer *peer;
+    char *domain;
+} idleconn;
+
+static void
+free_idleconn(void *data)
+{
+    idleconn *idle = data;
+    safe_free(idle->domain)
+}
+
+CBDATA_TYPE(idleconn);
+
+static void
+fwdConnectIdleDone(int fd, int status, void *data)
+{
+    idleconn *idle = data;
+    idle->peer->stats.idle_opening--;
+    if (fd >= 0)
+	pconnPush(fd, idle->peer->name, idle->peer->http_port, idle->domain, NULL, 0);
+    cbdataFree(idle);
+}
+
+static void
+fwdConnectIdleTimeout(int fd, void *data)
+{
+    idleconn *idle = data;
+    debug(17, 2) ("fwdConnectIdleTimeout: FD %d: '%s'\n", fd, idle->peer->name);
+    idle->peer->stats.idle_opening--;
+    comm_close(fd);
+    cbdataFree(idle);
+}
+
+static void
+openIdleConn(peer * peer, const char *domain, struct in_addr outgoing, unsigned short tos, int ctimeout)
+{
+    int fd = comm_openex(SOCK_STREAM,
+	IPPROTO_TCP,
+	outgoing,
+	0,
+	COMM_NONBLOCKING,
+	tos,
+	peer->name);
+    idleconn *idle;
+    if (fd < 0) {
+	debug(50, 4) ("openIdleConn: %s\n", xstrerror());
+	return;;
+    }
+    CBDATA_INIT_TYPE_FREECB(idleconn, free_idleconn);
+    idle = cbdataAlloc(idleconn);
+    idle->peer = peer;
+    cbdataLock(idle->peer);
+    idle->domain = xstrdup(domain);
+    /*
+     * stats.conn_open is used to account for the number of
+     * connections that we have open to the peer, so we can limit
+     * based on the max-conn option.  We need to increment here,
+     * even if the connection may fail.
+     */
+    peer->stats.conn_open++;
+    peer->stats.idle_opening++;
+    comm_add_close_handler(fd, fwdPeerClosed, peer);
+    commSetTimeout(fd,
+	ctimeout,
+	fwdConnectIdleTimeout,
+	idle);
+    commConnectStart(fd, peer->host, peer->http_port, fwdConnectIdleDone, idle);
+}
+
 static struct in_addr
 aclMapAddr(acl_address * head, aclCheck_t * ch)
 {
     acl_address *l;
     struct in_addr addr;
+    aclChecklistCacheInit(ch);
     for (l = head; l; l = l->next) {
 	if (aclMatchAclList(l->acl_list, ch))
 	    return l->addr;
@@ -414,6 +477,7 @@ static int
 aclMapTOS(acl_tos * head, aclCheck_t * ch)
 {
     acl_tos *l;
+    aclChecklistCacheInit(ch);
     for (l = head; l; l = l->next) {
 	if (aclMatchAclList(l->acl_list, ch))
 	    return l->tos;
@@ -427,9 +491,6 @@ getOutgoingAddr(request_t * request)
     aclCheck_t ch;
     memset(&ch, '\0', sizeof(aclCheck_t));
     if (request) {
-	ch.src_addr = request->client_addr;
-	ch.my_addr = request->my_addr;
-	ch.my_port = request->my_port;
 	ch.request = request;
     }
     return aclMapAddr(Config.accessList.outgoing_address, &ch);
@@ -441,9 +502,6 @@ getOutgoingTOS(request_t * request)
     aclCheck_t ch;
     memset(&ch, '\0', sizeof(aclCheck_t));
     if (request) {
-	ch.src_addr = request->client_addr;
-	ch.my_addr = request->my_addr;
-	ch.my_port = request->my_port;
 	ch.request = request;
     }
     return aclMapTOS(Config.accessList.outgoing_tos, &ch);
@@ -468,6 +526,7 @@ fwdConnectStart(void *data)
 #if LINUX_TPROXY
     struct in_tproxy itp;
 #endif
+    int idle = -1;
 
     assert(fs);
     assert(fwdState->server_fd == -1);
@@ -478,6 +537,8 @@ fwdConnectStart(void *data)
 	port = fs->peer->http_port;
 	if (fs->peer->options.originserver)
 	    domain = fwdState->request->host;
+	else
+	    domain = "*";
 	ctimeout = fs->peer->connect_timeout > 0 ? fs->peer->connect_timeout
 	    : Config.Timeout.peer_connect;
     } else {
@@ -512,15 +573,16 @@ fwdConnectStart(void *data)
 	fwdState->request->pinned_connection = NULL;
 	fwdState->servers = fs->next;
 	fwdServerFree(fs);
-	fwdConnectStart(fwdState);
+	fwdRestart(fwdState);
 	return;
     }
 #if LINUX_TPROXY
     if (fd == -1 && fwdState->request->flags.tproxy)
-	fd = pconnPop(name, port, domain, &fwdState->request->client_addr, 0);
+	fd = pconnPop(name, port, domain, &fwdState->request->client_addr, 0, NULL);
 #endif
-    if (fd == -1)
-	fd = pconnPop(name, port, domain, NULL, 0);
+    if (fd == -1) {
+	fd = pconnPop(name, port, domain, NULL, 0, &idle);
+    }
     if (fd != -1) {
 	if (fwdCheckRetriable(fwdState)) {
 	    debug(17, 3) ("fwdConnectStart: reusing pconn FD %d\n", fd);
@@ -530,12 +592,25 @@ fwdConnectStart(void *data)
 		fwdState->origin_tries++;
 	    comm_add_close_handler(fd, fwdServerClosed, fwdState);
 	    if (fs->peer)
-		hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
+		hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->name);
 	    else if (Config.onoff.log_ip_on_direct && fs->code == HIER_DIRECT)
 		hierarchyNote(&fwdState->request->hier, fs->code, fd_table[fd].ipaddr);
 	    else
 		hierarchyNote(&fwdState->request->hier, fs->code, name);
-	    fwdConnectDone(fd, COMM_OK, fwdState);
+	    if (fs->peer && idle >= 0 && idle < fs->peer->idle) {
+		debug(17, 3) ("fwdConnectStart: Opening idle connetions for '%s'\n",
+		    fs->peer->name);
+		outgoing = getOutgoingAddr(fwdState->request);
+		tos = getOutgoingTOS(fwdState->request);
+		debug(17, 3) ("fwdConnectStart: got addr %s, tos %d\n",
+		    inet_ntoa(outgoing), tos);
+		idle += fs->peer->stats.idle_opening;
+		while (idle < fs->peer->idle) {
+		    openIdleConn(fs->peer, domain, outgoing, tos, ctimeout);
+		    idle++;
+		}
+	    }
+	    fwdDispatch(fwdState);
 	    return;
 	} else {
 	    /* Discard the persistent connection to not cause
@@ -551,6 +626,8 @@ fwdConnectStart(void *data)
     outgoing = getOutgoingAddr(fwdState->request);
     tos = getOutgoingTOS(fwdState->request);
 
+    fwdState->request->out_ip = outgoing;
+
     debug(17, 3) ("fwdConnectStart: got addr %s, tos %d\n",
 	inet_ntoa(outgoing), tos);
     fd = comm_openex(SOCK_STREAM,
@@ -588,7 +665,7 @@ fwdConnectStart(void *data)
 	fwdConnectTimeout,
 	fwdState);
     if (fs->peer) {
-	hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->host);
+	hierarchyNote(&fwdState->request->hier, fs->code, fs->peer->name);
     } else {
 #if LINUX_TPROXY
 	if (fwdState->request->flags.tproxy) {
@@ -621,6 +698,16 @@ fwdConnectStart(void *data)
     commConnectStart(fd, host, port, fwdConnectDone, fwdState);
 }
 
+static void
+fwdRestart(void *data)
+{
+    FwdState *fwdState = data;
+    if (fwdState->servers)
+	fwdConnectStart(fwdState);
+    else
+	fwdStartFail(fwdState);
+}
+
 static void
 fwdStartComplete(FwdServer * servers, void *data)
 {
@@ -639,7 +726,7 @@ fwdStartFail(FwdState * fwdState)
 {
     ErrorState *err;
     debug(17, 3) ("fwdStartFail: %s\n", storeUrl(fwdState->entry));
-    err = errorCon(ERR_CANNOT_FORWARD, HTTP_SERVICE_UNAVAILABLE, fwdState->request);
+    err = errorCon(ERR_CANNOT_FORWARD, HTTP_GATEWAY_TIMEOUT, fwdState->request);
     err->xerrno = errno;
     fwdFail(fwdState, err);
     fwdStateFree(fwdState);
@@ -653,9 +740,10 @@ fwdDispatch(FwdState * fwdState)
     StoreEntry *entry = fwdState->entry;
     ErrorState *err;
     int server_fd = fwdState->server_fd;
+    FwdServer *fs = fwdState->servers;
     debug(17, 3) ("fwdDispatch: FD %d: Fetching '%s %s'\n",
 	fwdState->client_fd,
-	RequestMethodStr[request->method],
+	RequestMethods[request->method].str,
 	storeUrl(entry));
     /*
      * Assert that server_fd is set.  This is to guarantee that fwdState
@@ -667,6 +755,11 @@ fwdDispatch(FwdState * fwdState)
     assert(entry->ping_status != PING_WAITING);
     assert(entry->lock_count);
     EBIT_SET(entry->flags, ENTRY_DISPATCHED);
+    fd_note(server_fd, storeUrl(fwdState->entry));
+    fd_table[server_fd].uses++;
+    if (fd_table[server_fd].uses == 1 && fs->peer)
+	peerConnectSucceded(fs->peer);
+    fwdState->request->out_ip = fd_table[server_fd].local_addr;
     netdbPingSite(request->host);
     entry->mem_obj->refresh_timestamp = squid_curtime;
     if (fwdState->servers && (p = fwdState->servers->peer)) {
@@ -692,9 +785,6 @@ fwdDispatch(FwdState * fwdState)
 	case PROTO_FTP:
 	    ftpStart(fwdState);
 	    break;
-	case PROTO_WAIS:
-	    waisStart(fwdState);
-	    break;
 	case PROTO_CACHEOBJ:
 	case PROTO_INTERNAL:
 	case PROTO_URN:
@@ -703,6 +793,7 @@ fwdDispatch(FwdState * fwdState)
 	case PROTO_WHOIS:
 	    whoisStart(fwdState);
 	    break;
+	case PROTO_WAIS:	/* not implemented */
 	default:
 	    debug(17, 1) ("fwdDispatch: Cannot retrieve '%s'\n",
 		storeUrl(entry));
@@ -741,6 +832,10 @@ fwdReforward(FwdState * fwdState)
 	debug(17, 3) ("fwdReforward: No, ENTRY_FWD_HDR_WAIT isn't set\n");
 	return 0;
     }
+    if (fwdState->request->flags.pinned) {
+	debug(17, 3) ("fwdReforward: No, Connection oriented authentication is used\n");
+	return 0;
+    }
     if (fwdState->n_tries > 9)
 	return 0;
     if (fwdState->origin_tries > 1)
@@ -787,8 +882,10 @@ fwdStartPeer(peer * p, StoreEntry * e, request_t * r)
     fwdState->server_fd = -1;
     fwdState->request = requestLink(r);
     fwdState->start = squid_curtime;
+    fwdState->orig_entry_flags = e->flags;
     storeLockObject(e);
-    EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
+    if (!fwdState->request->flags.pinned)
+	EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
     storeRegisterAbort(e, fwdAbort, fwdState);
     peerAddFwdServer(&peer, p, HIER_DIRECT);
     fwdStartComplete(peer, fwdState);
@@ -798,7 +895,6 @@ void
 fwdStart(int fd, StoreEntry * e, request_t * r)
 {
     FwdState *fwdState;
-    aclCheck_t ch;
     int answer;
     ErrorState *err;
     /*
@@ -810,15 +906,10 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
 	/*      
 	 * Check if this host is allowed to fetch MISSES from us (miss_access)
 	 */
-	memset(&ch, '\0', sizeof(aclCheck_t));
-	ch.src_addr = r->client_addr;
-	ch.my_addr = r->my_addr;
-	ch.my_port = r->my_port;
-	ch.request = r;
-	answer = aclCheckFast(Config.accessList.miss, &ch);
+	answer = aclCheckFastRequest(Config.accessList.miss, r);
 	if (answer == 0) {
 	    err_type page_id;
-	    page_id = aclGetDenyInfoPage(&Config.denyInfoList, AclMatchedName);
+	    page_id = aclGetDenyInfoPage(&Config.denyInfoList, AclMatchedName, 1);
 	    if (page_id == ERR_NONE)
 		page_id = ERR_FORWARDING_DENIED;
 	    err = errorCon(page_id, HTTP_FORBIDDEN, r);
@@ -834,7 +925,7 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
 #endif
     if (shutting_down) {
 	/* more yuck */
-	err = errorCon(ERR_SHUTTING_DOWN, HTTP_SERVICE_UNAVAILABLE, r);
+	err = errorCon(ERR_SHUTTING_DOWN, HTTP_GATEWAY_TIMEOUT, r);
 	errorAppendEntry(e, err);
 	return;
     }
@@ -860,6 +951,7 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
     fwdState->server_fd = -1;
     fwdState->request = requestLink(r);
     fwdState->start = squid_curtime;
+    fwdState->orig_entry_flags = e->flags;
 
 #if LINUX_TPROXY
     /* If we need to transparently proxy the request
@@ -870,7 +962,8 @@ fwdStart(int fd, StoreEntry * e, request_t * r)
 #endif
 
     storeLockObject(e);
-    EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
+    if (!fwdState->request->flags.pinned)
+	EBIT_SET(e->flags, ENTRY_FWD_HDR_WAIT);
     storeRegisterAbort(e, fwdAbort, fwdState);
     peerSelect(r, e, fwdStartComplete, fwdState);
 }
@@ -1007,6 +1100,24 @@ fwdComplete(FwdState * fwdState)
 	if (fwdState->server_fd > -1)
 	    fwdUnregister(fwdState->server_fd, fwdState);
 	storeEntryReset(e);
+	/* HACK WARNING: This fiddling with the flags really
+	 * should be done in the store layer, but current
+	 * design does not allow it to be done proper in a
+	 * sane manner.
+	 * A sign that we have pushed the design of everything
+	 * going via a single StoreEntry per request a bit too far.
+	 */
+	if (EBIT_TEST(e->flags, ENTRY_NEGCACHED)) {
+	    storeSetPrivateKey(e);
+	    EBIT_CLR(e->flags, ENTRY_NEGCACHED);
+	    EBIT_CLR(e->flags, KEY_EARLY_PUBLIC);
+	}
+	if (EBIT_TEST(e->flags, RELEASE_REQUEST)) {
+	    EBIT_CLR(e->flags, RELEASE_REQUEST);
+	    EBIT_CLR(e->flags, KEY_EARLY_PUBLIC);
+	}
+	if (EBIT_TEST(fwdState->orig_entry_flags, ENTRY_CACHABLE))
+	    EBIT_SET(e->flags, ENTRY_CACHABLE);
 	fwdStartComplete(fwdState->servers, fwdState);
     } else {
 	debug(17, 3) ("fwdComplete: not re-forwarding status %d\n",
@@ -1115,7 +1226,7 @@ fwdLog(FwdState * fwdState)
 	(int) current_time.tv_sec,
 	(int) current_time.tv_usec / 1000,
 	fwdState->last_status,
-	RequestMethodStr[fwdState->request->method],
+	RequestMethods[fwdState->request->method].str,
 	fwdState->request->canonical);
 }
 
